from pathlib import Path
from typing import *  # type: ignore

import gitignore_parser

__all__ = [
    "PathMatch",
]


class PathMatch:
    """
    Takes a list of paths to include / exclude and provides a method to check
    whether a given path should be included or excluded.

    This class uses the same syntax as `.gitignore` files, with the exceptions
    that a match is considered positive, while `.gitignore` files consider a
    match negative.
    """

    def __init__(
        self,
        base_dir: Path,
        *,
        rules: Iterable[str] = tuple(),
    ) -> None:
        self._base_dir = base_dir.resolve()
        self._rules: list[gitignore_parser.IgnoreRule] = []

        for rule in rules:
            self.add_rule(rule)

    def add_rule(self, line: str) -> None:
        """
        Add a rule to the include list.
        """
        # Parse the rule
        line = line.rstrip("\n")
        rule = gitignore_parser.rule_from_pattern(
            line,
            base_path=self._base_dir,
            source=(None, len(self._rules)),
        )

        # Empty line / comment
        if not rule:
            return

        self._rules.append(rule)

    def match(self, path: Path) -> bool:
        """
        Given a path, return whether the given path matches any of the rules.
        """
        for rule in reversed(self._rules):
            if rule.match(path):
                return not rule.negation

        return False



from dataclasses import KW_ONLY
from typing import final

from .. import deprecations
from .fundamental_component import FundamentalComponent

__all__ = [
    "CodeBlock",
]


@final
@deprecations.parameters_renamed({"display_controls": "show_controls"})
class CodeBlock(FundamentalComponent):
    """
    Displays source code with syntax highlighting.

    The `CodeBlock` component displays source code with syntax highlighting,
    similar to how it would appear in a code editor or a markdown code block.

    If no language is specified, Rio will try to guess the language
    automatically. To disable syntax highlighting, set the language to `"text"`.

    ## Attributes

    `code`: The markdown-formatted text to display.

    `language`: The default language to use for syntax highlighting. If `None`,
        Rio will try to guess the language automatically.

    `show_controls`: Whether to display additional controls in addition to the
        source code itself. This includes a button to copy the code to the
        clipboard and a label for the language.
    """

    code: str

    _: KW_ONLY

    language: str | None = None
    show_controls: bool = True


CodeBlock._unique_id = "CodeBlock-builtin"



from __future__ import annotations

from dataclasses import KW_ONLY
from pathlib import Path
from typing import Literal, Union, final

from uniserde import JsonDoc

from .. import color, deprecations, fills, icons
from .fundamental_component import FundamentalComponent

__all__ = [
    "Icon",
]


_IconFill = Union[
    "fills.SolidFill",
    "fills.LinearGradientFill",
    "fills.ImageFill",
    "color.ColorSet",
    Literal["dim"],
]


@final
class Icon(FundamentalComponent):
    """
    Displays one of many pre-bundled icons.

    Icons are a great way to add polish to your app. A good icon can help your
    users understand your app and immediately recognize what a component does.

    Rio includes hundreds of free icons, allowing you to easily add them to your
    app without having to find or create your own. The `Icon` component displays
    one of these icons.

    Note that unlike most components in Rio, the `Icon` component does not have
    a `natural` size, since icons can be easily be scaled to fit any space.
    Because of this, `Icon` defaults to a width and height of 1.3, which is a
    great size when mixing icons with text.

    Icon names are in the format `set-name/icon-name:variant`. Rio already ships
    with the `material` icon set, which contains icons in the style of Google's
    Material Design. You can browse all available icons in Rio's dev tools. (The
    dev tools sidebar is visible on the right-hand-side when running your
    project using `rio run`.)

    The set name and variant can be omitted. If no set name is specified, it
    defaults to `material`. If no variant is specified, the default version of
    the icon, i.e. no variant, is used.


    ## Attributes

    `icon`: The name of the icon to display, in the format
        `icon-set/name:variant`. You can browse all available icons in Rio's
        dev tools sidebar.

    `fill`: The color scheme of the icon. The text color is used if no fill is
        specified.


    ## Examples

    This minimal example will display the icon named "castle" from the
    "material" icon set:

    ```python
    rio.Icon("material/castle")
    ```

    You can also specify the color, width and height of the icon:

    ```python
    rio.Icon(
        "material/castle",
        fill=rio.Color.from_hex("ff0000"),
        height=2.5,
        width=2.5,
    )
    ```
    """

    icon: str
    _: KW_ONLY
    fill: _IconFill

    @staticmethod
    @deprecations.deprecated(since="0.9.1", replacement=icons.register_icon_set)
    def register_icon_set(
        set_name: str,
        icon_set_archive_path: Path,
    ) -> None:
        icons.register_icon_set(set_name, icon_set_archive_path)

    @staticmethod
    @deprecations.deprecated(since="0.9.1", replacement=icons.register_icon)
    def register_single_icon(
        icon_source: Path,
        set_name: str,
        icon_name: str,
        variant_name: str | None = None,
    ) -> None:
        icons.register_icon(icon_source, set_name, icon_name, variant_name)

    def __init__(
        self,
        icon: str,
        *,
        fill: _IconFill = "keep",
        key: str | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        width: float | Literal["grow"] = 1.3,
        height: float | Literal["grow"] = 1.3,
        align_x: float | None = None,
        align_y: float | None = None,
    ):
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width=width,
            height=height,
            align_x=align_x,
            align_y=align_y,
        )

        self.icon = icon
        self.fill = fill

    def _custom_serialize(self) -> JsonDoc:
        # Serialize the fill. This isn't automatically handled because it's a
        # Union.
        fill = self.fill
        if not isinstance(fill, str):
            fill = self.session._serialize_fill(fill)

        # Serialize
        return {
            "fill": fill,
        }


Icon._unique_id = "Icon-builtin"



from __future__ import annotations

from dataclasses import KW_ONLY
from typing import final

from uniserde import JsonDoc

import rio

from .fundamental_component import FundamentalComponent

__all__ = ["ScrollTarget"]


@final
class ScrollTarget(FundamentalComponent):
    """
    Allows browsers to scroll to a specific component via URL fragment.

    `ScrollTarget` is a container which can be referenced by a URL fragment,
    which allows browsers to scroll to it when the page is loaded. For example,
    if your website contains a `ScrollTarget` with the ID `"my-section"`, then a
    browser visiting `https://your.website/#my-section` will immediately scroll
    it into view.

    When the mouse cursor is hovering over the `ScrollTarget`, it displays a `¶`
    symbol on its right side. Users can click this symbol to copy the URL to
    their clipboard. This symbol can be customized or removed through the
    `copy_button_content` parameter.


    ## Attributes

    `id`: The ID of the `ScrollTarget`. This must be unique among all
        `ScrollTarget`s on the page.

    `content`: The child component to display inside the `ScrollTarget`.

    `copy_button_content`: The text or component to use as the "Copy URL to
        clipboard" button. Setting this to `None` removes the button entirely.

    `copy_button_spacing`: The amount of empty space between the `content` and
        `copy_button_content`.


    ## Examples

    A minimal example of `ScrollTarget` displaying a heading:

    ```python
    rio.ScrollTarget(
        id="chapter-1",
        content=rio.Text("Chapter 1", style="heading1"),
    )
    ```

    ## Metadata

    `experimental`: True
    """

    id: str
    content: rio.Component | None = None
    _: KW_ONLY
    copy_button_content: str | rio.Component | None = "¶"
    copy_button_spacing: float = 0.5

    def _custom_serialize(self) -> JsonDoc:
        button_content = self.copy_button_content

        return {
            "copy_button_content": button_content._id
            if isinstance(button_content, rio.Component)
            else None,
            "copy_button_text": button_content
            if isinstance(button_content, str)
            else None,
        }


ScrollTarget._unique_id = "ScrollTarget-builtin"



class RioDeprecationWarning(Warning):
    """
    The user used/did something that's deprecated.
    """


class RioPotentialMistakeWarning(Warning):
    """
    The user did something sketchy that's probably a bad idea.
    """



from dataclasses import dataclass
from typing import *  # type: ignore

import uniserde

# Never import * from typing_extensions! It breaks `Any` on 3.10, preventing
# users from connecting. Ask me how I know.
from typing_extensions import Self

__all__ = ["ComponentLayout", "InitialClientMessage"]


@dataclass
class ComponentLayout:
    """
    Stores information about a component's layout. This includes the position
    and size that was actually allocated, rather than just requested.
    """

    # The component's position relative to the top-left corner of the viewport
    left_in_viewport: float
    top_in_viewport: float

    # The component's position relative to the top-left corner of the parent
    left_in_parent: float
    top_in_parent: float

    # How much space the component has requested
    natural_width: float
    natural_height: float

    # How much space the component was actually allocated
    allocated_width: float
    allocated_height: float

    # Aligned components are only given as much space as they need. This is how
    # much space the component was allocated prior to alignment.
    allocated_width_before_alignment: float
    allocated_height_before_alignment: float

    # Information about the parent. This can be useful in explaining the layout
    parent_id: int

    parent_natural_width: float
    parent_natural_height: float

    parent_allocated_width: float
    parent_allocated_height: float


@dataclass
class InitialClientMessage(uniserde.Serde):
    # The URL the client used to connect to the website. This can be quite
    # different from the URLs we see in FastAPI requests, because proxies like
    # nginx can alter it. For example, the client may be connecting via https
    # while rio only sees http.
    url: str

    # Don't annotate this as JsonDoc because uniserde doesn't support unions
    user_settings: dict[str, Any]

    prefers_light_theme: bool

    # List of RFC 5646 language codes. Most preferred first. May be empty!
    preferred_languages: list[str]

    # The names for all months, starting with January
    month_names_long: tuple[
        str,
        str,
        str,
        str,
        str,
        str,
        str,
        str,
        str,
        str,
        str,
        str,
    ]

    # The names of all weekdays, starting with monday
    day_names_long: tuple[
        str,
        str,
        str,
        str,
        str,
        str,
        str,
    ]

    # The format string for dates, as used by Python's `strftime`
    date_format_string: str

    # IANA timezone
    timezone: str

    # Separators for number rendering
    decimal_separator: str
    thousands_separator: str

    # Window Information
    window_width: float
    window_height: float

    @classmethod
    def from_defaults(
        cls,
        *,
        url: str,
        user_settings: uniserde.JsonDoc = {},
    ) -> Self:
        """
        Convenience method for creating default settings when they don't really
        matter: unit tests, crawlers, etc.
        """
        return cls(
            url=url,
            user_settings=user_settings,
            prefers_light_theme=True,
            preferred_languages=["en-US"],
            month_names_long=(
                "January",
                "February",
                "March",
                "April",
                "May",
                "June",
                "July",
                "August",
                "September",
                "October",
                "November",
                "December",
            ),
            day_names_long=(
                "Monday",
                "Tuesday",
                "Wednesday",
                "Thursday",
                "Friday",
                "Saturday",
                "Sunday",
            ),
            date_format_string="%Y-%m-%d",
            timezone="America/New_York",
            decimal_separator=".",
            thousands_separator=",",
            window_width=1920,
            window_height=1080,
        )



from __future__ import annotations

import dataclasses
import enum
from dataclasses import KW_ONLY, dataclass, is_dataclass
from typing import *  # type: ignore

import rio

from . import component, number_input, switch

__all__ = ["AutoForm"]


def prettify_name(name: str) -> str:
    parts = name.split("_")
    return " ".join(p.title() for p in parts)


@dataclass
class AutoFormChangeEvent:
    field_name: str
    value: Any


class AutoForm(component.Component):
    """
    TODO

    ## Metadata

    public: False
    """

    value: Any
    _: KW_ONLY
    on_change: rio.EventHandler[[AutoFormChangeEvent]] = None

    def __post_init__(self) -> None:
        # Make sure the passed value is a dataclass
        if not is_dataclass(self.value):
            raise TypeError(
                f"The value to `AutoForm` must be a dataclass, not `{type(self.value)}`"
            )

    async def _update_value(self, field_name: str, value: Any) -> None:
        # Update the value
        setattr(self, field_name, value)

        # Trigger the event
        await self.call_event_handler(
            self.on_change,
            AutoFormChangeEvent(
                field_name=field_name,
                value=value,
            ),
        )

    def _build_input_field(
        self,
        field_name: str,
        field_type: type,
    ) -> rio.Component:
        # Get sensible type information
        origin = get_origin(field_type)
        field_args = get_args(field_type)
        field_type = field_type if origin is None else origin
        del origin

        # Grab the value
        value = getattr(self.value, field_name)

        # `bool` -> `Switch`
        if field_type is bool:
            return switch.Switch(
                is_on=value,
                on_change=lambda e: self._update_value(field_name, e.is_on),
            )

        # `int` -> `NumberInput`
        if field_type is int:
            return number_input.NumberInput(
                value=value,
                decimals=0,
                on_change=lambda e: self._update_value(field_name, e.value),
            )

        # `float` -> `NumberInput`
        if field_type is float:
            return number_input.NumberInput(
                value=value,
                on_change=lambda e: self._update_value(field_name, e.value),
            )

        # `str` -> `TextInput`
        if field_type is str:
            return rio.TextInput(
                text=value,
                on_change=lambda e: self._update_value(field_name, e.text),
            )

        # `Literal` or `Enum` -> `Dropdown`
        if field_type is Literal or issubclass(field_type, enum.Enum):
            if field_type is Literal:
                mapping = {str(a): a for a in field_args}
            else:
                field_type = cast(Type[enum.Enum], field_type)
                mapping = {prettify_name(f.name): f.value for f in field_type}

            return rio.Dropdown(
                mapping,
                selected_value=value,
                on_change=lambda e: self._update_value(field_name, e.value),
            )

        # Unsupported type
        raise TypeError(
            f"AutoForm does not support fields of type `{field_type}`"
        )

    def build(self) -> rio.Component:
        grid = rio.Grid(
            row_spacing=0.5,
            column_spacing=1,
        )

        # Iterate over the class' fields and build the input fields
        for ii, field in enumerate(dataclasses.fields(self.value)):
            field_py_name = field.name
            field_type = field.type
            field_nicename = prettify_name(field_py_name)

            # Skip fields that are not supported
            try:
                input_component = self._build_input_field(
                    field_py_name, field_type
                )
            except TypeError:
                continue

            # Add a label
            grid.add(
                rio.Text(text=field_nicename, align_x=0),
                ii,
                0,
            )

            # And a input field
            grid.add(input_component, ii, 1)

        return grid



class NoFileSelectedError(Exception):
    pass


class AssetError(Exception):
    """
    Raised when an error occurs related to assets.

    Signifies that some operation related to assets has failed. E.g. trying to
    access a nonexistent asset.
    """

    def __init__(self, message: str) -> None:
        super().__init__(message)

    @property
    def message(self) -> str:
        """
        A human-readable message describing the problem.
        """
        return self.args[0]


class NavigationFailed(Exception):
    """
    Raised when navigation to a page fails.

    This exception is raised when attempting to navigate to a page, but the
    navigation fails for some reason. This could happen, for example, because a
    page guard throws an exception.

    Note that navigating to a nonexistent pages is not an error, as `PageViews`
    will simply display their fallback in that case. Thus this exception will
    not be raised in that case.
    """


class ClipboardError(Exception):
    """
    Exception raised for errors related to clipboard operations.
    """

    def __init__(self, message: str) -> None:
        super().__init__(message)

    @property
    def message(self) -> str:
        return self.args[0]



from __future__ import annotations

from dataclasses import KW_ONLY
from typing import *  # type: ignore

import rio

from .component import Component
from .fundamental_component import FundamentalComponent
from .progress_circle import ProgressCircle

__all__ = [
    "Button",
    "IconButton",
]


CHILD_MARGIN_X = 1.0
CHILD_MARGIN_Y = 0.3


@final
class Button(Component):
    """
    A clickable button.

    The `Button` component allows the user to trigger an action by clicking on
    it. You can use it to trigger a function call, navigate to a different page,
    or perform any other action you'd like.


    ## Attributes

    `content`: The text or child component to display inside of the button.

    `icon`: The name of an icon to display on the button, in the form
        "set/name:variant". See the `Icon` component for details of how icons
        work in Rio.

    `shape`: The shape of the button. This can be one of:

        - `"pill"`: A rectangle where the left and right sides are completely
            round.
        - `"rounded"`: A rectangle with rounded corners.
        - `"rectangle"`: A rectangle with sharp corners.

    `style`: Controls the button's appearance. This can be one of:

        - `"major"`: A highly visible button with bold visuals.
        - `"minor"`: A less visible button that doesn't stand out.
        - `"plain"`: A button with no background or border. Use this to blend
            unimportant buttons into the background.

    `color`: The color scheme to use for the button.

    `is_sensitive`: Whether the button should respond to user input.

    `is_loading`: Whether the button should display a loading indicator. Use
        this to indicate to the user that an action is currently running.

    `on_press`: Triggered when the user clicks on the button.


    ## Examples

    This code creates a button with the caption "Click me!":

    ```python
    rio.Button(
        "Click me!",
        on_press=lambda: print("Button pressed!"),
    )
    ```

    Icons are an easy way to make your app more visually appealing. Here's a
    button with an icon:

    ```python
    rio.Button(
        "Click me!",
        icon="material/mouse",
        on_press=lambda: print("Button pressed!"),
    )
    ```

    You can even put other components inside of the button. Here's a button with
    a progress bar that slowly fills up as you click it:

    ```python
    class ProgressButton(rio.Component):
        clicks: int = 0

        def _on_button_press(self) -> None:
            self.clicks += 1

        def build(self) -> rio.Component:
            return rio.Button(
                rio.Column(
                    rio.Text("Click repeatedly to fill up the progress bar"),
                    rio.ProgressBar(self.clicks / 10, width=15, height=1),
                ),
                on_press=self._on_button_press,
            )
    ```
    """

    content: str | rio.Component = ""
    _: KW_ONLY
    icon: str | None = None
    shape: Literal["pill", "rounded", "rectangle"] = "pill"
    style: Literal["major", "minor", "plain"] = "major"
    color: rio.ColorSet = "keep"
    is_sensitive: bool = True
    is_loading: bool = False
    on_press: rio.EventHandler[[]] = None

    def build(self) -> rio.Component:
        # Prepare the child
        if self.is_loading:
            if self.color in ("keep", "secondary"):
                progress_color = "primary"
            else:
                progress_color = "secondary"

            child = ProgressCircle(
                size=1.5,
                align_x=0.5,
                margin_x=CHILD_MARGIN_Y,
                margin_y=CHILD_MARGIN_Y,
                color=progress_color,
            )
        elif isinstance(self.content, Component):
            child = rio.Container(
                self.content,
                margin_x=CHILD_MARGIN_Y,
                margin_y=CHILD_MARGIN_Y,
            )
        else:
            children = []
            text = self.content.strip()
            n_children = (self.icon is not None) + bool(text)

            if self.icon is not None:
                children.append(
                    rio.Icon(
                        self.icon,
                        width=1.4,
                        height=1.4,
                        margin_x=CHILD_MARGIN_X if n_children == 1 else None,
                        margin_y=CHILD_MARGIN_Y if n_children == 1 else None,
                    )
                )

            if text:
                children.append(
                    rio.Text(
                        text,
                        justify="center",
                        margin_x=CHILD_MARGIN_X if n_children == 1 else None,
                        margin_y=CHILD_MARGIN_Y if n_children == 1 else None,
                        style=(
                            rio.TextStyle(font_weight="bold")
                            if self.style == "major"
                            else "text"
                        ),
                    )
                )

            if len(children) == 1:
                child = children[0]
            else:
                child = rio.Row(
                    *children,
                    spacing=0.6,
                    margin_x=CHILD_MARGIN_X,
                    margin_y=CHILD_MARGIN_Y,
                    align_x=0.5,
                )

        # Delegate to a HTML Component
        return _ButtonInternal(
            on_press=self.on_press,
            content=child,
            shape=self.shape,
            style=self.style,
            color=self.color,
            is_sensitive=self.is_sensitive,
            is_loading=self.is_loading,
            width=8 if isinstance(self.content, str) else "natural",
            height=2.2,
        )

    def __str__(self) -> str:
        if isinstance(self.content, str):
            content = f"text:{self.content!r}"
        else:
            content = f"content:{self.content._id}"

        return f"<Button id:{self._id} {content}>"


@final
class IconButton(Component):
    """
    # IconButton

    A round, clickable button with an icon.

    The `IconButton` component allows the user to trigger an action by clicking
    on it. You can use it to trigger a function call, navigate to a different
    page, or perform other actions.

    It is similar to the `Button` component, but it is specifically designed to
    display an icon, and it has a round shape.


    ## Attributes

    `icon`: The name of an icon to display on the button, in the form
            `"set/name:variant"`. See the `Icon` component for details of how
            icons work in Rio.

    `style`: Controls the button's appearance. This can be one of:

        - `major`: A highly visible button with bold visuals.
        - `minor`: A less visible button that blends into the background.
        - `plain`: A button with no background or border. Use this to make
                        the button look like a link.

    `color`: The color scheme to use for the button.

    `is_sensitive`: Whether the button should respond to user input.

    `size`: The size of the button. This is the diameter of the button in
            font-size units.

    `on_press`: Triggered when the user clicks on the button.


    ## Examples

    This minimal example will simply display a `IconButton` with a castle icon:

    ```python
    rio.IconButton(icon="material/castle")
    ```

    `IconButton`s are commonly used to trigger actions. You can easily achieve this by
    adding a function call to `on_press`:

    ```python
    class MyComponent(rio.Component):
        def on_press_button(self) -> None:
            print("Icon button pressed!")

        def build(self) -> rio.Component:
            return rio.IconButton(
                icon="material/castle",
                on_press=self.on_press_button,
            )
    ```

    `IconButton`s are commonly used to trigger actions. You can easily achieve
    this by adding a function call to `on_press`. You can use a function call to
    update the banner text signaling that the button was pressed:

    ```python
    class MyComponent(rio.Component):
        banner_text: str = ""

        def on_press_button(self) -> None:
            self.banner_text = "Icon button pressed!"

        def build(self) -> rio.Component:
            return rio.Column(
                rio.Banner(
                    text=self.banner_text,
                    style="info",
                ),
                rio.IconButton(
                    icon="material/castle",
                    on_press=self.on_press_button,
                ),
                spacing=1,
            )
    ```
    """

    icon: str
    _: KW_ONLY
    style: Literal["major", "minor", "plain"]
    color: rio.ColorSet
    is_sensitive: bool
    size: float
    on_press: rio.EventHandler[[]]

    def __init__(
        self,
        icon: str,
        *,
        style: Literal["major", "minor", "plain"] = "major",
        color: rio.ColorSet = "keep",
        is_sensitive: bool = True,
        on_press: rio.EventHandler[[]] = None,
        key: str | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        size: float = 3.7,
        align_x: float | None = None,
        align_y: float | None = None,
    ):
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width="natural",
            height="natural",
            align_x=align_x,
            align_y=align_y,
        )

        self.icon = icon
        self.size = size
        self.style = style
        self.color = color
        self.is_sensitive = is_sensitive
        self.on_press = on_press

    def build(self) -> rio.Component:
        return _ButtonInternal(
            on_press=self.on_press,
            content=rio.Icon(
                self.icon,
                height=self.size * 0.65,
                width=self.size * 0.65,
                align_x=0.5,
                align_y=0.5,
            ),
            shape="circle",
            style=self.style,
            color=self.color,
            is_sensitive=self.is_sensitive,
            is_loading=False,
            width=self.size,
            height=self.size,
            # Make sure the button has a square aspect ratio
            align_x=0.5,
            align_y=0.5,
        )

    def _get_debug_details(self) -> dict[str, Any]:
        result = super()._get_debug_details()

        # `width` & `height` are replaced with `size`
        del result["width"]
        del result["height"]

        return result


class _ButtonInternal(FundamentalComponent):
    _: KW_ONLY
    on_press: rio.EventHandler[[]]
    content: rio.Component
    shape: Literal["pill", "rounded", "rectangle", "circle"]
    style: Literal["major", "minor", "plain"]
    color: rio.ColorSet
    is_sensitive: bool
    is_loading: bool

    async def _on_message(self, msg: Any) -> None:
        # Parse the message
        assert isinstance(msg, dict), msg
        assert msg["type"] == "press", msg

        msg_type: str = msg["type"]
        assert isinstance(msg_type, str), msg_type

        # Is the button sensitive?
        if not self.is_sensitive or self.is_loading:
            return

        # Trigger the press event
        await self.call_event_handler(self.on_press)

        # Refresh the session
        await self.session._refresh()


_ButtonInternal._unique_id = "Button-builtin"



logo = "rio/logo"
logo_color = "rio/logo:color"
logo_fill = "rio/logo:fill"
logo_and_text_horizontal = "rio/logo-and-text-horizontal"
logo_and_text_horizontal_color = "rio/logo-and-text-horizontal:color"
logo_and_text_horizontal_fill = "rio/logo-and-text-horizontal:fill"
logo_and_text_vertical = "rio/logo-and-text-vertical"
logo_and_text_vertical_color = "rio/logo-and-text-vertical:color"
logo_and_text_vertical_fill = "rio/logo-and-text-vertical:fill"



from __future__ import annotations

import random
from dataclasses import KW_ONLY
from datetime import date

import rio

from .. import components as comps
from .component import Component

__all__ = [
    "DateInput",
]


class DateInput(Component):
    """
    Allows the user to pick a date from a calendar.

    DateInputs are similar in appearance to `TextInput` and `NumberInput`, but
    allow the user to pick a date from a calendar, rather than text or number.
    When pressed, a calendar will pop-up, allowing the user to select a date.

    This makes for a compact component, which still allows the user to visually
    select a date.

    For a larger component which permanently displays a calendar, consider using
    `rio.Calendar`.


    ## Attributes

    `value`: The currently selected date.

    `label`: A short text to display next to the input field.

    `on_change`: Triggered whenever the user selects a new date.


    ## Examples

    Here's a simple example that allows the user to select a data and displays
    it back to them:

    ```python
    from datetime import date

    class MyComponent(rio.Component):
        value: date = date.today()

        def build(self) -> rio.Component:
            return rio.Column(
                rio.DateInput(
                    # In order to retrieve a value from the component, we'll
                    # use an attribute binding. This way our own value will
                    # be updated whenever the user changes the value.
                    value=self.bind().value,
                    label="Pick a Date",
                ),
                rio.Text(f"You've selected: {self.value}"),
            )
    ```

    Alternatively you can also attach an event handler to react to changes. This
    is a little more verbose, but allows you to run arbitrary code when the user
    picks a new date:

    ```python
    from datetime import date

    class MyComponent(rio.Component):
        value: date = date.today()

        def on_value_change(self, event: rio.DateChangeEvent):
            # This function will be called whenever the input's value
            # changes. We'll display the new value in addition to updating
            # our own attribute.
            self.value = event.value
            print(f"You've selected: {self.value}")

        def build(self) -> rio.Component:
            return rio.DateInput(
                value=self.value,
                label="Pick a Date",
                on_change=self.on_value_change,
            )
    ```


    ## Metadata

    `experimental`: True
    """

    value: date

    _: KW_ONLY

    label: str = ""
    on_change: rio.EventHandler[rio.DateChangeEvent] = None

    _is_open: bool = False

    async def _on_value_change(self, event: rio.DateChangeEvent) -> None:
        # Close the date picker
        self._is_open = False

        # Chain the event handler
        await self.call_event_handler(self.on_change, event)

    def _on_toggle_open(self, _: rio.PressEvent) -> None:
        self._is_open = not self._is_open

    def _on_close(self) -> None:
        self._is_open = False

    def build(self) -> rio.Component:
        return rio.Popup(
            # Place a fake textbox. It's only used for styling and displaying
            # the label, if any
            anchor=rio.Stack(
                rio.TextInput(
                    label=self.label,
                    text=self.value.strftime(self.session._date_format_string),
                    is_sensitive=False,
                ),
                # This suppresses any clicks on the text input and opens the
                # calender instead
                rio.MouseEventListener(
                    rio.Rectangle(
                        content=rio.Icon(
                            "material/calendar-today:fill",
                            fill="dim",
                            height=1.5,
                            margin_right=0.5,
                            margin_bottom=0.5,
                            align_x=1,
                            align_y=1,
                        ),
                        fill=rio.Color.TRANSPARENT,
                        ripple=True,
                        cursor=rio.CursorStyle.POINTER,
                    ),
                    on_press=self._on_toggle_open,
                ),
                width=9,
            ),
            # Display a calendar so the user can pick a date
            content=rio.Column(
                comps.Calendar(
                    value=self.bind().value,
                    on_change=self._on_value_change,
                    # Use a fresh calendar each time, so it displays the
                    # currently selected value each time the popup opens
                    key=str(random.random()),
                ),
                rio.Button(
                    "Cancel",
                    style="plain",
                    on_press=self._on_close,
                ),
                # spacing=0.5,
                margin=1,
            ),
            color="neutral",
            position="center",
            alignment=0.5,
            is_open=self._is_open,
        )



from typing import *  # type: ignore

import rio.components.class_container
import rio.components.dev_tools_connector

from . import (
    deploy_page,
    docs_page,
    icons_page,
    project_page,
    theme_picker_page,
    tree_page,
)


class DevToolsSidebar(rio.Component):
    selected_page: (
        Literal[
            "project",
            "tree",
            "docs",
            "deploy",
        ]
        | None
    ) = None

    def get_selected_page(self) -> rio.Component | None:
        REGULAR_PAGE_WIDTH = 22
        WIDE_PAGE_WIDTH = 32

        # Nothing selected
        if self.selected_page is None:
            return None

        # Project
        if self.selected_page == "project":
            return project_page.ProjectPage(
                width=REGULAR_PAGE_WIDTH,
            )

        # Tree
        if self.selected_page == "tree":
            return tree_page.TreePage(
                width=WIDE_PAGE_WIDTH,
            )

        # Icons
        if self.selected_page == "icons":
            return icons_page.IconsPage(
                width=WIDE_PAGE_WIDTH,
            )

        # Theme
        if self.selected_page == "theme":
            return theme_picker_page.ThemePickerPage(
                width=WIDE_PAGE_WIDTH,
            )

        # Docs
        if self.selected_page == "docs":
            return docs_page.DocsPage(
                width=REGULAR_PAGE_WIDTH,
            )

        # Deploy
        if self.selected_page == "deploy":
            return deploy_page.DeployPage(
                width=REGULAR_PAGE_WIDTH,
            )

        # Anything else / TODO
        return rio.Text(
            f"TODO: {self.selected_page}",
            justify="center",
            margin=2,
            width=REGULAR_PAGE_WIDTH,
        )

    def build(self) -> rio.Component:
        return rio.Rectangle(
            # Make sure everything has a background, otherwise the component
            # highlighter will be visible behind this component
            fill=self.session.theme.background_color,
            content=rio.Row(
                # Big fat line to separate the dev tools from the rest of the page
                rio.Rectangle(
                    width=0.3,
                    fill=self.session.theme.primary_palette.background,
                ),
                # Currently active page
                rio.components.class_container.ClassContainer(
                    rio.Switcher(self.get_selected_page()),
                    classes=[
                        "rio-switcheroo-neutral",
                        "rio-dev-tools-background",
                    ],
                ),
                # Navigation
                rio.Column(
                    rio.SwitcherBar(
                        names=[
                            # "Project",
                            "Tree",
                            "Icons",
                            "Theme",
                            # "Docs",
                            "Deploy",
                        ],
                        icons=[
                            # "rio/logo",
                            "material/view-quilt",
                            "material/emoji-people",
                            "material/palette",
                            # "material/library-books",
                            "material/rocket-launch",
                        ],
                        values=[
                            # "project",
                            "tree",
                            "icons",
                            "theme",
                            # "docs",
                            "deploy",
                        ],
                        allow_none=True,
                        orientation="vertical",
                        spacing=2,
                        color="primary",
                        selected_value=self.bind().selected_page,
                        margin=0.2,
                    ),
                    rio.Spacer(),
                    rio.components.dev_tools_connector.DevToolsConnector(),
                ),
            ),
        )



from __future__ import annotations

from dataclasses import KW_ONLY, dataclass
from typing import Literal, TypeVar, final

from uniserde import JsonDoc

import rio.docs

from .fundamental_component import FundamentalComponent

__all__ = [
    "Revealer",
    "RevealerChangeEvent",
]

T = TypeVar("T")


@final
@rio.docs.mark_constructor_as_private
@dataclass
class RevealerChangeEvent:
    """
    Holds information regarding a revealer change event.

    This is a simple dataclass that stores useful information for when the user
    opens or closes a revealer. You'll typically receive this as argument in
    `on_change` events.

    ## Attributes

    `is_open`: The new `is_open` state of the `Revealer`.
    """

    is_open: bool


@final
class Revealer(FundamentalComponent):
    """
    A component that can be used to hide and reveal content.

    The `Revealer` is a versatile component that can be used to hide and reveal
    content. It can be used to create collapsible sections, or to hide and
    reveal content based on user input.

    If you specify a `header`, the revealer will automatically display a
    pressable header. Pressing the header will toggle the visibility of its
    content. If you don't specify a header, the revealer will only display its
    content, and must be expanded and collapsed programmatically.


    ## Attributes

    `header`: The header of the `Revealer`. If `None`, the `Revealer` will be
        hidden by default.

    `content`: The content to display when the `Revealer` is open.

    `header_style`: The style of the header. Can be one of `"heading1"`,
        `"heading2"`, `"heading3"`, or `"text"`.

    `is_open`: Whether the `Revealer` is open or not.

    `on_change`: An event handler that is called when the `Revealer` is opened
        or closed. The event handler receives a `RevealerChangeEvent` as input.


    ## Examples

    A simple `Revealer` that displays "Hello" when opened:

    ```python
    rio.Revealer(header="Click to reveal", content=rio.Text("Hello"))
    ```

    A `Revealer` is a component that hides content, in this case a TextInput,
    until it's opened:

    ```python
    class MyComponent(rio.Component):
        def build(self) -> rio.Component:
            return rio.Revealer(
                header="Click to Reveal",
                content=rio.TextInput(
                    text="Hey there!",
                ),
                header_style="heading2",
            )
    ```
    """

    header: str | None
    content: rio.Component
    _: KW_ONLY
    header_style: (
        Literal["heading1", "heading2", "heading3", "text"] | rio.TextStyle
    ) = "text"
    is_open: bool = False
    on_change: rio.EventHandler[RevealerChangeEvent] = None

    def _validate_delta_state_from_frontend(self, delta_state: JsonDoc) -> None:
        if not set(delta_state) <= {"is_open"}:
            raise AssertionError(
                f"Frontend tried to change `{type(self).__name__}` state: {delta_state}"
            )

        if "is_open" in delta_state and self.header is None:
            raise AssertionError(
                f"Frontend tried to set `Revealer.is_open` even though it has no `header`"
            )

    async def _call_event_handlers_for_delta_state(
        self, delta_state: JsonDoc
    ) -> None:
        # Trigger on_change event
        try:
            new_value = delta_state["is_open"]
        except KeyError:
            pass
        else:
            assert isinstance(new_value, bool), new_value
            await self.call_event_handler(
                self.on_change,
                RevealerChangeEvent(new_value),
            )

    def _custom_serialize(self) -> JsonDoc:
        # Serialization doesn't handle unions. Hence the custom serialization
        # here
        if isinstance(self.header_style, str):
            header_style = self.header_style
        else:
            header_style = self.header_style._serialize(self.session)

        return {
            "header_style": header_style,
        }


Revealer._unique_id = "Revealer-builtin"



from __future__ import annotations

import abc

from uniserde import Jsonable

import rio

__all__ = ["SelfSerializing"]


class SelfSerializing(abc.ABC):
    """
    Properties with types that inherit from `SelfSerializing` will be serialized
    when sent to the client.
    """

    @abc.abstractmethod
    def _serialize(self, sess: rio.Session) -> Jsonable:
        raise NotImplementedError



from __future__ import annotations

from typing import *  # type: ignore

import rio


# <component>
class Navbar(rio.Component):
    """
    A navbar with a fixed position and responsive width.
    """

    # Make sure the navbar will be rebuilt when the app navigates to a different
    # page. While Rio automatically detects state changes and rebuilds
    # components as needed, navigating to other pages is not considered a state
    # change, since it's not stored in the component.
    #
    # Instead, we can use Rio's `on_page_change` event to trigger a rebuild of
    # the navbar when the page changes.
    @rio.event.on_page_change
    async def on_page_change(self) -> None:
        # Rio comes with a function specifically for this. Whenever Rio is
        # unable to detect a change automatically, use this function to force a
        # refresh.
        await self.force_refresh()

    def build(self) -> rio.Component:
        # Which page is currently active? This will be used to highlight the
        # correct navigation button.
        #
        # `active_page_instances` contains the same `rio.Page` instances that
        # you've passed the app during creation. Since multiple pages can be
        # active at a time (e.g. /foo/bar/baz), this is a list.
        active_page = self.session.active_page_instances[0]
        active_page_url_segment = active_page.page_url

        # The navbar should appear above all other components. This is easily
        # done by using a `rio.Overlay` component.
        return rio.Overlay(
            rio.Row(
                rio.Spacer(),
                # Use a card for visual separation
                rio.Rectangle(
                    content=rio.Row(
                        # Links can be used to navigate to other pages and
                        # external URLs. You can pass either a simple string, or
                        # another component as their content.
                        rio.Link(
                            rio.IconButton(
                                "rio/logo",
                                style="plain",
                                size=2.5,
                            ),
                            "/",
                        ),
                        # This spacer will take up any superfluous space,
                        # effectively pushing the subsequent buttons to the
                        # right.
                        rio.Spacer(),
                        # By sticking buttons into a `rio.Link`, we can easily
                        # make the buttons navigate to other pages, without
                        # having to write an event handler. Notice how there is
                        # no Python function called when the button is clicked.
                        rio.Link(
                            rio.Button(
                                "News",
                                icon="material/news",
                                style=(
                                    "major"
                                    if active_page_url_segment == "news-page"
                                    else "plain"
                                ),
                            ),
                            "/news-page",
                        ),
                        # Same game, different button
                        rio.Link(
                            rio.Button(
                                "About",
                                icon="material/info",
                                style=(
                                    "major"
                                    if active_page_url_segment == "about-page"
                                    else "plain"
                                ),
                            ),
                            "/about-page",
                        ),
                        spacing=1,
                        margin=1,
                    ),
                    fill=self.session.theme.neutral_color,
                    corner_radius=self.session.theme.corner_radius_medium,
                    shadow_radius=0.8,
                    shadow_color=self.session.theme.shadow_color,
                    shadow_offset_y=0.2,
                ),
                rio.Spacer(),
                # Proportions are an easy way to make the navbar's size relative
                # to the screen. This assigns 5% to the first spacer, 90% to the
                # navbar, and 5% to the second spacer.
                proportions=(0.5, 9, 0.5),
                # Overlay assigns the entire screen to its child component.
                # Since the navbar isn't supposed to take up all space, assign
                # an alignment.
                align_y=0,
                margin=2,
            )
        )


# </component>



from __future__ import annotations

from dataclasses import KW_ONLY
from typing import Literal, final

import rio

from .fundamental_component import FundamentalComponent

__all__ = ["ScrollContainer"]


@final
class ScrollContainer(FundamentalComponent):
    """
    Displays a scroll bar if its content grows too large.

    `ScrollContainer` is a container which can display child components that
    would normally be too large to fit on the screen. It accepts and displays a
    single child and adds scroll bars if smaller than the child. If you can also
    force enable or disable scrollbars using the `scroll_x` and `scroll_y`
    properties.


    ## Attributes

    `content`: The child component to display inside the `ScrollContainer`.

    `scroll_x`: Controls horizontal scrolling. The default is `"auto"`, which
        means that a scroll bar will be displayed only if it is needed.
        `"always"` displays a scroll bar even if it isn't needed, and
        `"never"` disables horizontal scrolling altogether.

    `scroll_y`: Controls vertical scrolling. The default is `"auto"`, which
        means that a scroll bar will be displayed only if it is needed.
        `"always"` displays a scroll bar even if it isn't needed, and
        `"never"` disables vertical scrolling altogether.

    `initial_x`: The initial location of the scroll knob along the horizontal
        axis. Ranges from 0 (left) to 1 (right).

    `initial_y`: The initial location of the scroll knob along the vertical
        axis. Ranges from 0 (top) to 1 (bottom).

    `sticky_bottom`: If `True`, when the user has scrolled to the bottom and
        the content of the `ScrollContainer` grows larger, the scroll bar
        will automatically scroll to the bottom again.


    ## Examples

    A minimal example of `ScrollContainer` displaying an icon:

    ```python
    rio.ScrollContainer(
        content=rio.Icon("material/castle", width=50, height=50),
        height=10,
    )
    ```
    """

    content: rio.Component
    _: KW_ONLY
    scroll_x: Literal["never", "auto", "always"] = "auto"
    scroll_y: Literal["never", "auto", "always"] = "auto"
    initial_x: float = 0
    initial_y: float = 0
    sticky_bottom: bool = False


ScrollContainer._unique_id = "ScrollContainer-builtin"



import subprocess
import sys

import revel

DEV_BRANCH = "dev"
RELEASE_BRANCH = "main"


def main() -> None:
    revel.print("Running sanity checks...")

    ensure_branch(DEV_BRANCH)
    ensure_no_uncommitted_changes()
    ensure_up_to_date_with_remote()

    build_frontend()
    ensure_tests_pass()

    revel.print("Everything is in order.")
    make_new_release()


def ensure_branch(branch_name: str) -> None:
    process = subprocess.run(
        ["git", "branch", "--show-current"],
        check=True,
        capture_output=True,
        text=True,
    )
    if process.stdout.strip("\n") != branch_name:
        revel.fatal(f'You must checkout the "{branch_name}" branch')


def ensure_no_uncommitted_changes() -> None:
    process = subprocess.run(
        ["git", "status", "--porcelain", "--untracked-files=no"],
        check=True,
        capture_output=True,
        text=True,
    )
    if process.stdout.strip():
        revel.fatal("There are uncommitted changes")


def ensure_up_to_date_with_remote() -> None:
    subprocess.run(["git", "remote", "update"], check=True)
    process = subprocess.run(
        ["git", "status", "--porcelain", "--untracked-files=no"],
        check=True,
        capture_output=True,
        text=True,
    )
    if process.stdout.strip():
        revel.fatal("Local branch is not up-to-date with remote")


def build_frontend() -> None:
    subprocess.run(
        ["rye", "run", "build"],
        check=True,
    )


def ensure_tests_pass() -> None:
    process = subprocess.run(["rye", "test", "--", "-x", "--disable-warnings"])
    if process.returncode != 0:
        revel.fatal("The test suite does not pass")


def make_new_release() -> None:
    # Bump the version
    version_part_to_bump = revel.select(
        {
            "patch": "patch",
            "minor": "minor",
            "major": "major",
        },
        prompt="Which version do you want to bump?",
    )
    subprocess.run(
        [sys.executable, "-m", "hatch", "version", version_part_to_bump],
        check=True,
    )
    subprocess.run(["git", "commit", "--all", "-m", "bump version"], check=True)
    subprocess.run(["git", "push"], check=True)

    # Merge DEV_BRANCH into RELEASE_BRANCH and push
    subprocess.run(
        ["git", "fetch", ".", f"{DEV_BRANCH}:{RELEASE_BRANCH}"], check=True
    )
    subprocess.run(["git", "push", "-u", "origin", RELEASE_BRANCH], check=True)

    # Publish
    subprocess.run(["rye", "build", "--clean"], check=True)
    subprocess.run(["rye", "publish"], check=True)


if __name__ == "__main__":
    main()



from __future__ import annotations

import abc
import asyncio
import inspect
import json
import logging
import time
import traceback
import warnings
import weakref
from datetime import date
from pathlib import Path
from typing import *

import langcodes
import pytz
import starlette.datastructures

import rio
import rio.assets

from .. import (
    assets,
    data_models,
    language_info,
    routing,
    session,
    user_settings_module,
    utils,
)
from ..transports import (
    AbstractTransport,
    TransportClosedIntentionally,
    TransportInterrupted,
)

__all__ = ["AbstractAppServer"]


class AbstractAppServer(abc.ABC):
    def __init__(
        self,
        app: rio.App,
        *,
        running_in_window: bool,
        debug_mode: bool,
    ):
        self.app = app
        self.running_in_window = running_in_window
        self.debug_mode = debug_mode

        self._session_serve_tasks = dict[rio.Session, asyncio.Task[object]]()
        self._disconnected_sessions = dict[rio.Session, float]()

    @property
    def sessions(self) -> list[rio.Session]:
        return list(self._session_serve_tasks)

    async def _on_start(self) -> None:
        # Trigger the app's startup event
        #
        # This will be done blockingly, so the user can prepare any state before
        # connections are accepted. This is also why it's called before the
        # internal event.
        await self._call_on_app_start()

        # If running as a server, periodically clean up expired sessions
        if not self.running_in_window:
            asyncio.create_task(
                _periodically_clean_up_expired_sessions(weakref.ref(self)),
                name="Periodic session cleanup",
            )

    async def _on_close(self) -> None:
        # Close all sessions
        rio._logger.debug(
            f"App server shutting down; closing {len(self.sessions)} active session(s)"
        )

        results = await asyncio.gather(
            *(sess._close(close_remote_session=True) for sess in self.sessions),
            return_exceptions=True,
        )
        for result in results:
            if isinstance(result, BaseException):
                traceback.print_exception(
                    type(result), result, result.__traceback__
                )

        await self._call_on_app_close()

    async def _call_on_app_start(self) -> None:
        rio._logger.debug("Calling `on_app_start`")

        if self.app._on_app_start is None:
            return

        try:
            result = self.app._on_app_start(self.app)

            if inspect.isawaitable(result):
                await result

        # Display and discard exceptions
        except Exception:
            print("Exception in `on_app_start` event handler:")
            traceback.print_exc()

    async def _call_on_app_close(self) -> None:
        rio._logger.debug("Calling `on_app_close`")

        if self.app._on_app_close is None:
            return

        try:
            result = self.app._on_app_close(self.app)

            if inspect.isawaitable(result):
                await result

        # Display and discard exceptions
        except Exception:
            print("Exception in `_on_app_close` event handler:")
            traceback.print_exc()

    @abc.abstractmethod
    async def file_chooser(
        self,
        session: rio.Session,
        *,
        file_extensions: Iterable[str] | None = None,
        multiple: bool = False,
    ) -> utils.FileInfo | tuple[utils.FileInfo, ...]:
        raise NotImplementedError

    @abc.abstractmethod
    def url_for_user_asset(self, relative_asset_path: Path) -> rio.URL:
        """
        Returns the URL of an asset from the app's `assets_dir`. This must be a
        permalink, i.e. a URL that doesn't change if the server is restarted.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def weakly_host_asset(self, asset: assets.HostedAsset) -> str:
        """
        Hosts an asset as long as it is alive. Returns the asset's URL as a
        string.
        """
        raise NotImplementedError

    def host_asset_with_timeout(
        self, asset: assets.HostedAsset, timeout: float
    ) -> str:
        """
        Hosts an asset for a limited time. Returns the asset's URL as a string.
        """
        url = self.weakly_host_asset(asset)

        async def keep_alive():
            await asyncio.sleep(timeout)
            _ = asset

        asyncio.create_task(
            keep_alive(), name=f"Keep asset {asset} alive for {timeout} sec"
        )

        return url

    def _after_session_closed(self, session: rio.Session) -> None:
        """
        Called by `Session.close()`. Gives the server an opportunity to clean
        up.
        """
        # Stop the task that's listening for incoming messages
        task = self._session_serve_tasks.pop(session)
        task.cancel()

    async def create_session(
        self,
        initial_message: data_models.InitialClientMessage,
        *,
        transport: AbstractTransport,
        client_ip: str,
        client_port: int,
        http_headers: starlette.datastructures.Headers,
    ) -> rio.Session:
        """
        Creates a new session.

        ## Raises

        `NavigationFailed`: If a page guard crashes
        """
        # Normalize and deduplicate the languages
        preferred_languages: list[str] = []

        for language in initial_message.preferred_languages:
            try:
                language = langcodes.standardize_tag(language)
            except ValueError:
                continue

            if language not in preferred_languages:
                preferred_languages.append(language)

        if len(preferred_languages) == 0:
            preferred_languages.append("en-US")

        # Get locale information
        if len(initial_message.decimal_separator) != 1:
            logging.warning(
                f'Client sent invalid decimal separator "{initial_message.decimal_separator}". Using "." instead.'
            )
            initial_message.decimal_separator = "."

        if len(initial_message.thousands_separator) > 1:
            logging.warning(
                f'Client sent invalid thousands separator "{initial_message.thousands_separator}". Using "" instead.'
            )
            initial_message.thousands_separator = ""

        # There does not seem to be any good way to determine the first day of
        # the week in JavaScript. Look it up based on the preferred language.
        first_day_of_week = language_info.get_week_start_day(
            preferred_languages[0]
        )

        # Make sure the date format string is valid
        try:
            formatted_date = date(3333, 11, 22).strftime(
                initial_message.date_format_string
            )
        except ValueError:
            date_format_string_is_valid = False
        else:
            date_format_string_is_valid = (
                "33" in formatted_date
                and "11" in formatted_date
                and "22" in formatted_date
            )

        if not date_format_string_is_valid:
            logging.warning(
                f'Client sent invalid date format string "{initial_message.date_format_string}". Using "%Y-%m-%d" instead.'
            )
            initial_message.date_format_string = "%Y-%m-%d"

        # Parse the timezone
        try:
            timezone = pytz.timezone(initial_message.timezone)
        except pytz.UnknownTimeZoneError:
            logging.warning(
                f'Client sent unknown timezone "{initial_message.timezone}". Using UTC instead.'
            )
            timezone = pytz.UTC

        # Set the theme according to the user's preferences
        theme = self.app._theme
        if isinstance(theme, tuple):
            if initial_message.prefers_light_theme:
                theme = theme[0]
            else:
                theme = theme[1]

        # Prepare the initial URL. This will be exposed to the session as the
        # `active_page_url`, but overridden later once the page guards have been
        # run.
        initial_page_url = rio.URL(initial_message.url.lower())
        base_url = (
            initial_page_url.with_path("").with_query("").with_fragment("")
        )

        # Create the session
        sess = session.Session(
            app_server_=self,
            transport=transport,
            client_ip=client_ip,
            client_port=client_port,
            http_headers=http_headers,
            base_url=base_url,
            active_page_url=initial_page_url,
            timezone=timezone,
            preferred_languages=preferred_languages,
            month_names_long=initial_message.month_names_long,
            day_names_long=initial_message.day_names_long,
            date_format_string=initial_message.date_format_string,
            first_day_of_week=first_day_of_week,
            decimal_separator=initial_message.decimal_separator,
            thousands_separator=initial_message.thousands_separator,
            window_width=initial_message.window_width,
            window_height=initial_message.window_height,
            theme_=theme,
        )

        # Deserialize the user settings
        await sess._load_user_settings(initial_message.user_settings)

        # Add any remaining attachments
        for attachment in self.app.default_attachments:
            if not isinstance(attachment, user_settings_module.UserSettings):
                sess.attach(attachment)

        # Start listening for incoming messages. This should happen before
        # `on_session_start` is called, so that we don't deadlock in case
        # someone calls a method that requires a response from the client.
        self._session_serve_tasks[sess] = asyncio.create_task(
            self._serve_session(sess),
            name=f"`Session.serve` for session id `{id(sess)}`",
        )

        # Trigger the `on_session_start` event.
        #
        # Since this event is often used for important initialization tasks like
        # adding attachments, actually wait for it to finish before continuing.
        #
        # However, also don't run it too early, since this function expects a
        # (mostly) functioning session.
        #
        # TODO: Figure out which values are still missing, and expose them,
        #       expose placeholders, or document that they aren't available.
        start_time = time.monotonic()
        await sess._call_event_handler(
            self.app._on_session_start,
            sess,
            refresh=False,
        )
        duration = time.monotonic() - start_time

        if duration > 5:
            warnings.warn(
                f"Session startup was delayed for {duration:.0f} seconds by"
                f" `on_session_start`. If you have long-running operations that"
                f" don't have to finish before the session starts, you should"
                f" execute them in a background task."
            )

        # Run any page guards for the initial page. Throws a `NavigationFailed`
        # if a page guard crashed.
        (
            active_page_instances,
            active_page_url_absolute,
        ) = routing.check_page_guards(sess, initial_page_url)

        # Is this a page, or a full URL to another site?
        try:
            utils.make_url_relative(
                sess._base_url,
                active_page_url_absolute,
            )

        # This is an external URL. Navigate to it
        except ValueError:

            async def history_worker() -> None:
                await sess._evaluate_javascript(
                    f"""
                    window.location.href = {json.dumps(str(active_page_url_absolute))};
                    """
                )

            sess.create_task(history_worker(), name="navigate to external URL")

            # TODO: End the session? Abort initialization?

        # Set the initial page URL. When connecting to the server, all relevant
        # page guards execute. These may change the URL of the page, so the
        # client needs to take care to update the browser's URL to match the
        # server's.
        if active_page_url_absolute != initial_page_url:

            async def update_url_worker():
                js_page_url = json.dumps(str(active_page_url_absolute))
                await sess._evaluate_javascript(
                    f"""
                    console.trace("Updating browser URL to match the one modified by guards:", {js_page_url});
                    window.history.replaceState(null, "", {js_page_url});
                    """
                )

            sess.create_task(
                update_url_worker(),
                name="Update browser URL to match the one modified by guards",
            )

        # Update the session's active page and instances
        sess._active_page_instances = tuple(active_page_instances)
        sess._active_page_url = active_page_url_absolute

        # Apply the CSS for the chosen theme
        await sess._apply_theme(theme)

        # Send the first `updateComponentStates` message
        await sess._refresh()

        return sess

    async def _serve_session(self, sess: rio.Session) -> None:
        try:
            await sess.serve()
        except TransportClosedIntentionally:
            sess.close()
        except TransportInterrupted:
            # Connection was interrupted, mark the session as disconnected but
            # keep it alive for a while to see if the client reconnects
            sess._transport = None


async def _periodically_clean_up_expired_sessions(
    app_server_ref: weakref.ReferenceType[AbstractAppServer],
):
    LOOP_INTERVAL = 60 * 15
    SESSION_LIFETIME = 60 * 60

    while True:
        await asyncio.sleep(LOOP_INTERVAL)

        app_server = app_server_ref()
        if app_server is None:
            return

        now = time.monotonic()
        cutoff = now - SESSION_LIFETIME

        disconnected_sessions = list(app_server._disconnected_sessions.items())
        for sess, disconnect_time in disconnected_sessions:
            if disconnect_time < cutoff:
                sess.close()

        # Drop the reference to the app server
        app_server = None



from __future__ import annotations

from typing import *  # type: ignore

import rio


# <component>
class ChatSuggestionCard(rio.Component):
    """
    While the chat is empty, a placeholder is displayed. That placeholder
    contains several of these components, which are suggestions for the user to
    start a conversation.
    """

    icon: str
    text: str

    on_press: rio.EventHandler[str] = None

    async def _on_press(self) -> None:
        await self.call_event_handler(self.on_press, self.text)

    def build(self) -> rio.Component:
        # A suggestion is just an icon, text and button wrapped inside a card.
        return rio.Card(
            rio.Column(
                rio.Icon(self.icon),
                rio.Text(
                    self.text,
                    justify="center",
                    wrap=True,
                    height="grow",
                    align_y=0.5,
                ),
                rio.Button(
                    "Ask",
                    icon="material/navigate-next",
                    style="minor",
                    on_press=self._on_press,
                ),
                spacing=0.6,
                margin=1,
            ),
        )


# </component>



from collections.abc import Iterable

import imy.docstrings
import pytest

import rio.docs


def _create_tests():
    for obj, docs in rio.docs.find_documented_objects().items():
        if isinstance(docs, imy.docstrings.FunctionDocs):
            test_cls = _create_function_tests(docs)
        else:
            assert isinstance(obj, type)
            test_cls = _create_class_tests(obj, docs)

        globals()[test_cls.__name__] = test_cls


def _create_function_tests(docs: imy.docstrings.FunctionDocs) -> type:
    # If the function is a decorator, there's no need to document that it takes
    # a function/class as an argument
    if (
        docs.metadata.decorator
        and len(docs.parameters) == 1
        and docs.parameters[0].name == "handler"
    ):
        parameters = []
    else:
        parameters = docs.parameters

    class Tests:  # type: ignore
        def test_summary(self) -> None:
            assert docs.summary is not None, f"{docs.name} has no summary"

        def test_details(self) -> None:
            assert docs.details is not None, f"{docs.name} has no details"

        @pytest.mark.parametrize(
            "param",
            parameters,
            ids=[param.name for param in parameters],
        )
        def test_param_description(
            self, param: imy.docstrings.FunctionParameter
        ) -> None:
            assert (
                param.description is not None
            ), f"{docs.name}.{param.name} has no description"

    Tests.__name__ = f"Test_{docs.name}"
    return Tests


def _create_class_tests(cls: type, docs: imy.docstrings.ClassDocs) -> type:
    methods = [
        func
        for func in docs.functions
        if func.name != "__init__" or not issubclass(cls, rio.Component)
    ]

    methods_excluding_init = [
        func for func in methods if func.name != "__init__"
    ]

    # Components only need their constructor documented, attributes don't matter
    attributes = [] if issubclass(cls, rio.Component) else docs.attributes

    class Tests:
        def test_summary(self) -> None:
            assert docs.summary is not None, f"{cls.__name__} has no summary"

        def test_details(self) -> None:
            assert docs.details is not None, f"{cls.__name__} has no details"

        @parametrize_with_name("attr", attributes)
        def test_attribute_description(
            self, attr: imy.docstrings.ClassField
        ) -> None:
            assert (
                attr.description is not None
            ), f"{cls.__name__}.{attr.name} has no description"

        # __init__ methods don't need a summary
        @parametrize_with_name("method", methods_excluding_init)
        def test_method_summary(
            self, method: imy.docstrings.FunctionDocs
        ) -> None:
            assert (
                method.summary is not None
            ), f"{cls.__name__}.{method.name} has no summary"

        # __init__ methods don't need details
        @parametrize_with_name("method", methods_excluding_init)
        def test_method_details(
            self, method: imy.docstrings.FunctionDocs
        ) -> None:
            assert (
                method.details is not None
            ), f"{cls.__name__}.{method.name} has no details"

        @parametrize_with_name("method", methods)
        def test_method_parameters(
            self, method: imy.docstrings.FunctionDocs
        ) -> None:
            for param in method.parameters[1:]:
                assert (
                    param.description is not None
                ), f"Parameter {param.name!r} has no description"

    Tests.__name__ = f"Test{docs.name}"
    return Tests


def parametrize_with_name(
    param_name: str,
    docs: Iterable[
        imy.docstrings.FunctionDocs
        | imy.docstrings.ClassDocs
        | imy.docstrings.ClassField
        | imy.docstrings.FunctionParameter
    ],
):
    def decorator(func):
        pytest.mark.parametrize(
            param_name,
            docs,
            ids=[doc.name for doc in docs],
        )(func)

    return decorator


_create_tests()



from __future__ import annotations

from typing import *  # type: ignore

import rio


# <component>
class GeneratingResponsePlaceholder(rio.Component):
    """
    This component is displayed while the chatbot is generating a response. It
    provides feedback to the user that the app is working on their question.
    """

    def build(self) -> rio.Component:
        return rio.Row(
            rio.ProgressCircle(size=1.5),
            rio.Text(
                "Thinking...",
                justify="center",
            ),
            spacing=1,
        )


# </component>
